#include <stdint.h>
#include <time.h>
#include <string.h>

#include "ow_rom.h"
#include "ds1904.h"
#include "crc8.h"
#include "hw_profile.h"
#include "xprintf.h"

OWRTC owrtc;

OWRTC_RESULT initOWRTC(uint8_t * ROM)
{
    owrtc.valid = 0;

    if (crc8(ROM, 8) != 0) {
        return OWRTC_CRCERR;
    }

    if (ROM[0] != OW_FAMILY_CLOCK) {
        return OWRTC_NOTCLOCK;
    }

    if (!OWReset()) {
        return OWRTC_NOTFOUND;
    }

    memmove(ROM_NO, ROM, 8);
    memmove(owrtc.ROM, ROM, 8);

    if (!OWVerify()) {
        return OWRTC_NOTCLOCK;
    }

    owrtc.valid = 1;
    return OWRTC_OK;
}

OWRTC_RESULT read_clock(time_t * ptt)
{
    if (!owrtc.valid) {
        return OWRTC_NOTVALID;
    }
    if (!OWReset()) {
        return OWRTC_NOTFOUND;
    }

    memmove(ROM_NO, owrtc.ROM, 8);

#ifdef OW_SHOULD_VERIFY
    if (!OWVerify()) {
        return OWRTC_NOTCLOCK;
    }
#endif
    OWMatch();
    OWWriteByte(READ_CLOCK);

    owrtc.control_byte = OWReadByte(); /* control byte */
    if ((owrtc.control_byte & 0x3) != 0) {
        return OWRTC_CTRLBYTE;
    }

    owrtc.time_bytes[0] = OWReadByte();
    owrtc.time_bytes[1] = OWReadByte();
    owrtc.time_bytes[2] = OWReadByte();
    owrtc.time_bytes[3] = OWReadByte();

    OWReset();

    *ptt = owrtc.unixtime;
    return OWRTC_OK;
}

OWRTC_RESULT write_clock(time_t tt)
{
    if (!owrtc.valid) {
        return OWRTC_NOTVALID;
    }
    if (!OWReset()) {
        return OWRTC_NOTFOUND;
    }
    memmove(ROM_NO, owrtc.ROM, 8);
#ifdef OW_SHOULD_VERIFY
    if (!OWVerify()) {
        return OWRTC_NOTCLOCK;
    }
#endif

    OWMatch();
    OWWriteByte(WRITE_CLOCK);

    owrtc.unixtime = tt;

    OWWriteByte(owrtc.control_byte); /* control byte */

    OWWriteByte(owrtc.time_bytes[0]);
    OWWriteByte(owrtc.time_bytes[1]);
    OWWriteByte(owrtc.time_bytes[2]);
    OWWriteByte(owrtc.time_bytes[3]);

    OWReset();

    return OWRTC_OK;
}

OWRTC_RESULT read_clock_single(time_t * ptt)
{

    if (!OWReset()) {
        return OWRTC_NOTFOUND;
    }
    OWWriteByte(SKIP_ROM);
    OWWriteByte(READ_CLOCK);
    owrtc.control_byte = OWReadByte(); /* control byte */
    if ((owrtc.control_byte & 0x3) != 0) {
        return OWRTC_CTRLBYTE;
    }

    owrtc.time_bytes[0] = OWReadByte();
    owrtc.time_bytes[1] = OWReadByte();
    owrtc.time_bytes[2] = OWReadByte();
    owrtc.time_bytes[3] = OWReadByte();

    OWReset();

    *ptt = owrtc.unixtime;
    return OWRTC_OK;
}

OWRTC_RESULT write_clock_single(time_t tt)
{
    if (!OWReset()) {
        return OWRTC_NOTFOUND;
    }
    OWWriteByte(SKIP_ROM);
    OWWriteByte(WRITE_CLOCK);

    owrtc.unixtime = tt;
    owrtc.control_byte &= 0xFC;
    owrtc.control_byte |= 0xC;

    OWWriteByte(owrtc.control_byte); /* control byte */

    OWWriteByte(owrtc.time_bytes[0]);
    OWWriteByte(owrtc.time_bytes[1]);
    OWWriteByte(owrtc.time_bytes[2]);
    OWWriteByte(owrtc.time_bytes[3]);

    OWReset();

    return OWRTC_OK;
}

OWRTC_RESULT write_flags(uint8_t flags)
{
    time_t now;
    OWRTC_RESULT owres;
    owres = read_clock(&now);
    if (owres != OWRTC_OK) {
        return owres;
    }
    owrtc.control_byte |= (flags << 4);
    owres = write_clock(now);
    if (owres != OWRTC_OK) {
        return owres;
    }
    return OWRTC_OK;
}

OWRTC_RESULT read_flags(uint8_t* pflags)
{
    time_t dummy;
    OWRTC_RESULT owres;
    owres = read_clock(&dummy);
    if (owres != OWRTC_OK) {
        return owres;
    }
    /* 4b general purpose flags */
    *pflags = (owrtc.control_byte >> 4) & 0xF;
    return OWRTC_OK;
}

OWRTC_RESULT set_control_byte(uint8_t control_byte) {
    owrtc.control_byte = control_byte;
    return OWRTC_OK;
}