#include "hw_profile.h"

void config_osc(void) {
#if defined(__dsPIC33F__) || defined(__PIC24H__)
    uint8_t oscconh = 0b000;
#ifdef XT
    oscconh = 0b010;
#else
    oscconh = 0b000;
#endif

#ifdef PLL
    oscconh |= 0b1;
    // Fosc= Fin*M/(N1*N2), Fcy=Fosc/2
    CLKDIVbits.PLLPOST = (N1 / 2) - 1;
    PLLFBD = M - 2;
    CLKDIVbits.PLLPRE = N2 - 2;
#endif

    OSCTUN = 0; // Tune FRC oscillator, if FRC is used

    __builtin_write_OSCCONH(oscconh); // Initiate Clock Switch
    __builtin_write_OSCCONL(0x1); // Start clock switching

    // Wait for Clock switch to occur
    while (OSCCONbits.COSC != oscconh) {
        continue;
    }
#ifdef PLL
    // Wait for PLL to lock
    while (OSCCONbits.LOCK != 0b1) {
        continue;
    }
#endif
#elif defined(__PIC24F__)
    _RCDIV = 0;
#elif defined(__PIC24FK__)
	CLKDIV = 0x0;
#endif

}

#ifdef USE_CHECK_STARTUP

#include "xprintf.h"

/** \name Bitfield to struct conversion
 *  <a name="bitfieldMacros">Given</a> a bitfield struct representing an SFR,
 *  convert this to a word (uint16_t) or to the
 *  high and low bytes (uint8_t) of that word.
 *  Note: though this looks ugly, it introduces
 *  no unexpected compiler overhead at -O1. See
 *  \ref u16_INTTREGlast for an example.
 */
//@{
/// Convert a bitfield to a word (uint16_t).
#define BITS2WORD(sfrBitfield)  ( *((uint16_t*) &sfrBitfield) )
/// Return the low byte (as a uint8_t) of a bitfield.
#define BITS2BYTEL(sfrBitfield) ( ((uint8_t*)  &sfrBitfield)[0] )
/// Return the high byte (as a uint8_t) of a bitfield.
#define BITS2BYTEH(sfrBitfield) ( ((uint8_t*)  &sfrBitfield)[1] )
//@}


/** Persistent storage for an error message, typically
 *  set by \ref reportError and reported at reset by
 *  \ref printResetCause.
 */
static _PERSISTENT const char* sz_lastError;
/** Persistent storage for a timeout error, to be reported
 *  if a watchdog reset occurs.
 */
_PERSISTENT const char* sz_lastTimeoutError;

#ifndef __PIC24F__
/** Store a copy of the INTTREG register as a bitfield.
 *  This is not defined for the PIC24F, so work around
 *  with an \#ifdef. This is _PERSISTENT so that it
 *  survives the resets which occurs immeidately after
 *  the default interrupt handler \ref _DefaultInterrupt
 *  copies INTTREG to this variable.
 */
static _PERSISTENT INTTREGBITS INTTREGBITS_last;
/** Make the \ref INTTREGBITS_last also accessible as
 *  a word. This is like <code>uint16_t u16_INTTREGlast</code>
 *  except that INTTREGBITS_last and u16_INTTREGlast
 *  refer to the same data.
 */
#define u16_INTTREGlast BITS2WORD(INTTREGBITS_last)
#else
static uint16_t u16_INTTREGlast;
#endif

/** Report a critical error by recording a message
 *  in \ref sz_lastError then resetting the chip,
 *  assuming hat \ref printResetCause will be called
 *  during chip initialization.
 *  \param sz_errorMessage Error message to report.
 *  \see REPORT_ERROR
 */
void reportError(const char* sz_errorMessage) {
    //ignore if a previous error has already been triggerred
    if (sz_lastError == NULL) {
        sz_lastError = sz_errorMessage;
        asm("reset");
    }
}

/** Determines and prints the cause of a CPU reset. This should
 *  be called when the chip first starts up. For an example, see
 *  the \ref configBasic function.
 */
int printResetCause(void) {
    uint8_t u8_resetIdentified;

    u8_resetIdentified = 0;
    if (_SLEEP) {
        xputs("\n$$INFO,Device has been in sleep mode\n");
        _SLEEP = 0;
    }
    if (_IDLE) {
        xputs("\n$$INFO,Device has been in idle mode\n");
        _IDLE = 0;
    }
    xputs("\n$$INFO,Reset cause: ");
    if (_POR) {
        u8_resetIdentified = 2;
        xputs("Power-on.\n");
        _POR = 0;
        _BOR = 0; //clear both
        // Set the values below, which reset all the
        // error reporting variables to indicate that
        // no error has (yet) occurred.

        sz_lastError = NULL;
        sz_lastTimeoutError = NULL;
        u16_INTTREGlast = 0;
    } else {
        //non-POR
        if (_SWR) {
            xputs("Software Reset.\n");
            u8_resetIdentified = 1;
            _SWR = 0;
            _EXTR = 0; //also sets the EXTR bit
        }
        if (_WDTO) {
            u8_resetIdentified = 1;
            xputs("Watchdog Timeout: ");
            /*
            if (sz_lastTimeoutError != NULL) {
              xputs(sz_lastTimeoutError);
            }
             * */
            xputs("\n");
            _WDTO = 0;
            _EXTR = 0; //also sets the EXTR bit
        }
        if (_EXTR) {
            u8_resetIdentified = 1;
            xputs("MCLR assertion.\n");
            _EXTR = 0;
        }
        if (_BOR) {
            u8_resetIdentified = 2;
            xputs("Brown-out.\n");
            _BOR = 0;

        }
        if (_TRAPR) {
            u8_resetIdentified = 1;
            xputs("Trap Conflict.\n");
            _TRAPR = 0;
        }
        if (_IOPUWR) {
            u8_resetIdentified = 1;
            xputs("Illegal Condition.\n");
            _IOPUWR = 0;
        }
#ifdef _CM
        if (_CM) {
            u8_resetIdentified = 1;
            xputs("Configuration Mismatch.\n");
            _CM = 0;
        }
#endif
    } //end non-POR

    if (!u8_resetIdentified) {
        xputs("Unknown reset.\n");
    }
    if (sz_lastError != NULL) {
        xputs("Error trapped: ");
        xputs(sz_lastError);
        if (u16_INTTREGlast != 0) {
#if ( defined(__PIC24H__) || defined(__dsPIC33F__) )
            xprintf("Priority: %02X", INTTREGBITS_last.ILR);
            xprintf(", Vector number: %02X", INTTREGBITS_last.VECNUM);
#else
            xputs("Unknown priority/vector");
#endif
        }
        xputs("\n\n");
        sz_lastError = NULL;
        u16_INTTREGlast = 0;
    }


    return u8_resetIdentified;
}

/** Provide a default interrupt handler which records what
 *  interrupt was not handled then resets the chip. Typically,
 *  a call to \ref printResetCause during chip startup will then
 *  print the error.
 */
void _ISR __attribute__((no_auto_psv))_DefaultInterrupt(void) {
#ifndef __PIC24F__
    // Record the interrupt vector and priority of the
    // unhandled interrupt.
    u16_INTTREGlast = INTTREG;
#else
    // Any non-zero value causes reportError to report
    // this. This register doesn't exist on the PIC24F.
    u16_INTTREGlast = 1;
#endif
    reportError("Unhandled interrupt, ");
}
#else 
//void _ISRFAST __attribute__((no_auto_psv)) _DefaultInterrupt(void) {
//  asm("reset");
//}

int printResetCause(void) {
    return 0;
}
#endif
