#include "hw_profile.h"
#include "baudrate.h"
#include "uart1_i.h"
#include "fifo.h"

#define RX_BUFFER_SIZE 128
#define TX_BUFFER_SIZE 128

volatile uint16_t nUartRxLines = 0;

static volatile int TxRun; /* Tx running flag */
FIFO_BUFFER rx_uart1buf;
FIFO_BUFFER tx_uart1buf;
static uint8_t rx_buff[RX_BUFFER_SIZE];
static uint8_t tx_buff[TX_BUFFER_SIZE];

/* uart1 Rx interrupt ISR */
void __attribute__((interrupt, auto_psv)) _U1RXInterrupt(void)
{
    char d;
    d = (uint8_t) U1RXREG; /* Get received data */
#ifdef RT_CMD_U1
    if (d == '\n' || d == '\r') {
        nUartRxLines++;
    }
#endif
    FIFO_Put(&rx_uart1buf, d);
    _U1RXIF = 0; /* Clear Rx interrupt flag */
}

/* uart1 Tx interrupt ISR */
void __attribute__((interrupt, auto_psv)) _U1TXInterrupt(void)
{
    _U1TXIF = 0; /* Clear Tx interrupt flag */

    if (!FIFO_Empty(&tx_uart1buf)) { /* If any data is available, pop a uint8_t and send it. */
        U1TXREG = FIFO_Get(&tx_uart1buf); /* Send a uint8_t */
    } else { /* No data in the Tx FIFO */
        TxRun = 0; /* Stop transmission sequense */
    }
}

/* Check number of bytes in the Rx FIFO */
unsigned uart_test(void)
{
#ifdef RT_CMD_U1
    return nUartRxLines;
#else
    return FIFO_Count(&rx_uart1buf);
#endif
}

/* Check number of bytes in the Rx FIFO */
bool uart_empty(void)
{
    return FIFO_Empty(&rx_uart1buf);
}

/* Get a byte from Rx FIFO */
uint8_t uart_get(void)
{
    uint8_t d;

    while (FIFO_Empty(&rx_uart1buf)) {
        continue;
    }

    d = FIFO_Get(&rx_uart1buf);
#ifdef RT_CMD_U1
    if (d == '\n' || d == '\r') {
        if (nUartRxLines) {
            uint16_t u1rxie = _U1RXIE;
            _U1RXIE = 0;
            nUartRxLines--;
            _U1RXIE = u1rxie;
        }
    }
#endif
    return d;
}

/* Put a byte into Tx FIFO */
void uart_put(uint8_t d)
{
    while (!FIFO_Put(&tx_uart1buf, d)) {
        continue;
    }

    if (!TxRun) { /* If transmission sequense is not running, start the tx sequense */
        TxRun = 1;
        _U1TXIF = 1; /* Force trigger Tx interrupt */
    }
}

/* Initialize uart1 module */
void uart_init(void)
{
    /* Disable uart1 Tx/Rx interrupts */
    _U1RXIE = 0;
    _U1TXIE = 0;

    /* Clear Tx/Rx FIFOs */
    TxRun = 0;
    FIFO_Init(&rx_uart1buf, rx_buff, sizeof(rx_buff));
    FIFO_Init(&tx_uart1buf, tx_buff, sizeof(tx_buff));

    /* Initialize UART1 */
    U1BRG = BAUDRATEREG;
    U1MODE = 0;
    U1MODEbits.BRGH = BRGH_CONST;
    U1MODEbits.UARTEN = 1;
    U1STA = 0;
    U1STAbits.UTXEN = 1;

    /* Enable uart1 Tx/Rx interruptrs */
    _U1RXIP = 4;
    _U1RXIE = 1;
    _U1TXIE = 1;
}

void uart_clear(void)
{
    uint16_t u1rxie = _U1RXIE,
            u1txie = _U1TXIE;
    /* Clear Tx/Rx FIFOs */
    _U1RXIE = 0;
    _U1TXIE = 0;
    nUartRxLines = 0;
    /* Clear Tx/Rx FIFOs */
    TxRun = 0;
    FIFO_Init(&rx_uart1buf, rx_buff, sizeof(rx_buff));
    FIFO_Init(&tx_uart1buf, tx_buff, sizeof(tx_buff));
    _U1RXIE = u1rxie;
    _U1TXIE = u1txie;
}

void uart_shut(void)
{
    /* Disable UART1 Tx/Rx interrupts */
    _U1RXIE = 0;
    _U1TXIE = 0;

    U1MODE = 0;
    U1STA = 0;

    /* Clear Tx/Rx FIFOs */
    TxRun = 0;
    FIFO_Init(&rx_uart1buf, rx_buff, sizeof(rx_buff));
    FIFO_Init(&tx_uart1buf, tx_buff, sizeof(tx_buff));
}

void uart_init_br(uint32_t baudrate)
{
    /* Disable uart1 Tx/Rx interrupts */
    _U1RXIE = 0;
    _U1TXIE = 0;

    /* Clear Tx/Rx FIFOs */
    TxRun = 0;
    FIFO_Init(&rx_uart1buf, rx_buff, sizeof(rx_buff));
    FIFO_Init(&tx_uart1buf, tx_buff, sizeof(tx_buff));

    U1STA = 0;
    U1MODE = 0;
    /* Initialize UART1 */
    U1BRG = calc_bgr(baudrate, 1);
    U1MODEbits.BRGH = 1;
    U1MODEbits.UARTEN = 1;
    U1STAbits.UTXEN = 1;

    /* Enable uart1 Tx/Rx interruptrs */
    _U1RXIP = 0;
    _U1RXIE = 0;
    _U1TXIE = 1;
}
