#include "hw_profile.h"
#include "baudrate.h"
#include "uart2_i.h"
#include "fifo.h"


static volatile int TxRun; /* Tx running flag */
FIFO_BUFFER rx_uart2buf;
FIFO_BUFFER tx_uart2buf;
static uint8_t rx_buff[RX_BUFFER_SIZE];
static uint8_t tx_buff[TX_BUFFER_SIZE];

#ifdef USE_GPS

enum {
    GPS_NONE = 0,
    GPS_PREF = 1,
    GPS_CHK = 2
};

volatile uint16_t ngpslines = 0;
uint16_t gpsrx_state = GPS_NONE;
uint16_t cmdcount = 0;
uint16_t cmdnum = 1;

const uint16_t cmdlen = 5;
char *prefcmd[] = {"GPRMC"};
#endif

#ifdef RT_CMD_U2
volatile uint16_t nUart2RxLines = 0;
#endif

/* uart2 Rx interrupt ISR */
void __attribute__((interrupt, auto_psv)) _U2RXInterrupt(void)
{
    char d;
#ifdef USE_GPS
    int i;
    d = (uint8_t) U2RXREG; /* Get received data */
    _U2RXIF = 0; /* Clear Rx interrupt flag */
    _LATC6 = 1 - _LATC6;
    if (gpsrx_state == GPS_NONE) {
        if (d == '$') {
            gpsrx_state = GPS_PREF;
            cmdcount = 0;
        }
    } else if (gpsrx_state == GPS_PREF) {
        if (cmdcount < cmdlen) {
            for (i = 0; i < cmdnum; i++) {
                if (d == prefcmd[i][cmdcount]) {
                    break;
                }
            }
            if (i < cmdnum) { // prefix cmd0
                cmdcount++;
            } else { // another prefix
                gpsrx_state = GPS_NONE;
            }
        } else if (cmdcount >= cmdlen && d == ',') {
            cmdcount = 0;
            gpsrx_state = GPS_CHK;
            FIFO_Put(&rx_uart2buf, d);
        }
    } else if (gpsrx_state == GPS_CHK) {
        if (d == '\n') {
            ngpslines++;
            gpsrx_state = GPS_NONE;
        }
        FIFO_Put(&rx_uart2buf, d);
    }
#endif
    d = (uint8_t) U2RXREG; /* Get received data */

    _U2RXIF = 0; /* Clear Rx interrupt flag */
    FIFO_Put(&rx_uart2buf, d);
#ifdef RT_CMD_U2
    if (d == '\n' || d == '\r') {
        nUart2RxLines++;
    }
#endif
}

/* uart2 Tx interrupt ISR */
void __attribute__((interrupt, auto_psv)) _U2TXInterrupt(void)
{
    _U2TXIF = 0; /* Clear Tx interrupt flag */

    if (!FIFO_Empty(&tx_uart2buf)) { /* If any data is available, pop a uint8_t and send it. */
        U2TXREG = FIFO_Get(&tx_uart2buf); /* Send a uint8_t */
    } else { /* No data in the Tx FIFO */
        TxRun = 0; /* Stop transmission sequense */
    }
}

/* Check number of bytes in the Rx FIFO */
uint16_t uart2_test(void)
{
#ifdef RT_CMD_U2
    return nUart2RxLines;
#else
    return (uint16_t) FIFO_Empty(&tx_uart2buf);
#endif
}

/* Check number of bytes in the Rx FIFO */
bool uart2_empty(void)
{
    return FIFO_Empty(&tx_uart2buf);
}

/* Get a byte from Rx FIFO */
uint8_t uart2_get(void)
{
    uint8_t d;

    while (FIFO_Empty(&rx_uart2buf)) {
        continue;
    }

    d = FIFO_Get(&rx_uart2buf);
#ifdef RT_CMD_U2
    if (d == '\n' || d == '\r') {
        if (nUart2RxLines) {
            uint16_t u2rxie = _U2RXIE;
            _U2RXIE = 0;
            nUart2RxLines--;
            _U2RXIE = u2rxie;
        }
    }
#endif
    return d;
}

/* Put a byte into Tx FIFO */
void uart2_put(uint8_t d)
{
    while (!FIFO_Put(&tx_uart2buf, d)) {
        continue;
    }

    if (!TxRun) { /* If transmission sequense is not running, start the tx sequense */
        TxRun = 1;
        _U2TXIF = 1; /* Force trigger Tx interrupt */
    }
}

/* Initialize uart2 module */
void uart2_init(void)
{
    /* Disable uart21 Tx/Rx interrupts */
    _U2RXIE = 0;
    _U2TXIE = 0;

    /* Clear Tx/Rx FIFOs */
    TxRun = 0;
    FIFO_Init(&rx_uart2buf, rx_buff, sizeof(rx_buff));
    FIFO_Init(&tx_uart2buf, tx_buff, sizeof(tx_buff));

#ifdef USE_GPS
    ngpslines = 0;
    gpsrx_state = GPS_NONE;
    cmdcount = 0;
#endif

#ifdef RT_CMD_U2
    nUart2RxLines = 0;
#endif

    /* Initialize UART1 */
    U2BRG = BAUDRATEREG;
    U2MODE = 0;
    U2MODEbits.BRGH = BRGH_CONST;
    U2MODEbits.UARTEN = 1;
    U2STA = 0;
    U2STAbits.UTXEN = 1;

    /* Enable uart2 Tx/Rx interruptrs */
    _U2RXIP = 3;
    _U2RXIE = 1;
    _U2TXIE = 1;
}

void uart2_clear(void)
{
    uint16_t u2rxie = _U2RXIE,
            u2txie = _U2TXIE;
    /* Clear Tx/Rx FIFOs */
    _U2RXIE = 0;
    _U2TXIE = 0;
#ifdef USE_GPS
    ngpslines = 0;
    gpsrx_state = GPS_NONE;
    cmdcount = 0;
#endif

#ifdef RT_CMD_U2
    nUart2RxLines = 0;
#endif

    TxRun = 0;
    FIFO_Init(&rx_uart2buf, rx_buff, sizeof(rx_buff));
    FIFO_Init(&tx_uart2buf, tx_buff, sizeof(tx_buff));
    _U2RXIE = u2rxie;
    _U2TXIE = u2txie;
}

void uart2_shut(void)
{
    /* Disable UART1 Tx/Rx interrupts */
    _U2RXIE = 0;
    _U2TXIE = 0;

    U2MODE = 0;
    U2STA = 0;

#ifdef USE_GPS
    ngpslines = 0;
    gpsrx_state = GPS_NONE;
    cmdcount = 0;
#endif

#ifdef RT_CMD_U2
    nUart2RxLines = 0;
#endif

    TxRun = 0;
    FIFO_Init(&rx_uart2buf, rx_buff, sizeof(rx_buff));
    FIFO_Init(&tx_uart2buf, tx_buff, sizeof(tx_buff));
}

/* Initialize uart2 module */
void uart2_init_br(uint32_t baudrate)
{
    /* Disable uart21 Tx/Rx interrupts */
    _U2RXIE = 0;
    _U2TXIE = 0;

    /* Clear Tx/Rx FIFOs */
    TxRun = 0;
    FIFO_Init(&rx_uart2buf, rx_buff, sizeof(rx_buff));
    FIFO_Init(&tx_uart2buf, tx_buff, sizeof(tx_buff));

#ifdef USE_GPS
    ngpslines = 0;
    gpsrx_state = GPS_NONE;
    cmdcount = 0;
#endif

    U2STA = 0;
    U2MODE = 0;
    /* Initialize UART1 */
    U2BRG = calc_bgr(baudrate, 1);
    U2MODEbits.BRGH = 1;
    U2MODEbits.UARTEN = 1;
    U2STAbits.UTXEN = 1;

    /* Enable uart2 Tx/Rx interruptrs */
    _U2RXIP = 3;
    _U2RXIE = 1;
    _U2TXIE = 1;
}